#include <logging/log.h>
// #include <openthread/coap.h>
#include <openthread/link.h>
#include <openthread/platform/radio.h>
#include <openthread/thread.h>
#include <net/openthread.h>
#include <zephyr.h>
#include "thread.h"

LOG_MODULE_REGISTER(thread, LOG_LEVEL_DBG);

#define THREAD_CHANNEL 11
#define THREAD_PAN_ID 0xDEAD
#define THREAD_EXT_PAN_ID "56:a6:2b:e8:5e:4f:81:bc"
#define THREAD_NETWORK_KEY "00:11:22:33:44:55:66:77:88:99:aa:bb:cc:dd:ee:ff"
#define THREAD_NETWORK_NAME "Nikola"
#define THREAD_PASSPHRASE "N1K0LA"
#define THREAD_JOINER_TIMEOUT_MS 120

static bool bIsConnected = false;
static bool bIsCommissionerActive = false;

static void prvOtConnectCallback(struct k_work *pxItem);
static void prvOtDisconnectCallback(struct k_work *pxItem);
static void prvStartCommissioner(struct k_work *pxItem);
static void prvAddJoiner(struct k_work *pxItem);

K_WORK_DEFINE(xOnConnectWork, prvOtConnectCallback);
K_WORK_DEFINE(xOnDisconnectWork, prvOtDisconnectCallback);
K_WORK_DEFINE(xStartCommissionerWork, prvStartCommissioner);
K_WORK_DEFINE(xAddJoinerWork, prvAddJoiner);

static void prvOtStateChangeCallback(uint32_t ulFlags, void *pvContext)
{
	if (ulFlags & OT_CHANGED_THREAD_ROLE)
	{
		switch (otThreadGetDeviceRole(openthread_get_default_instance()))
		{
			case OT_DEVICE_ROLE_CHILD:
			case OT_DEVICE_ROLE_ROUTER:
			case OT_DEVICE_ROLE_LEADER:
				k_work_submit(&xOnConnectWork);
				bIsConnected = true;
				break;

			case OT_DEVICE_ROLE_DISABLED:
			case OT_DEVICE_ROLE_DETACHED:
			default:
				k_work_submit(&xOnDisconnectWork);
				bIsConnected = false;
				break;
		}
	}
}

static void prvOtCommissionerStateCallback(otCommissionerState xState, void *pvContext)
{
	switch (xState)
	{
		case OT_COMMISSIONER_STATE_DISABLED:
			bIsCommissionerActive = false;
			break;
		case OT_COMMISSIONER_STATE_PETITION:
			bIsCommissionerActive = false;
			break;
		case OT_COMMISSIONER_STATE_ACTIVE:
			bIsCommissionerActive = true;

			// TODO: Remove this later
			k_work_submit(&xAddJoinerWork);
			break;
		default:
			break;
	}
}

static void prvOtCommissionerJoinerCallback(otCommissionerJoinerEvent xEvent, const otJoinerInfo *pxJoinerInfo, const otExtAddress *xJoinerId, void *pvContext)
{
	switch (xEvent)
	{
		case OT_COMMISSIONER_JOINER_START:
			LOG_INF("Joiner Start");
			break;
		case OT_COMMISSIONER_JOINER_CONNECTED:
			LOG_INF("Joiner Connected");
			break;
		case OT_COMMISSIONER_JOINER_FINALIZE:
			LOG_INF("Joiner Finalize");
			break;
		case OT_COMMISSIONER_JOINER_END:
			LOG_INF("Joiner End");
			break;
		case OT_COMMISSIONER_JOINER_REMOVED:
			LOG_INF("Joiner Removed");
			break;
		default:
			break;
	}
}

static void prvOtConnectCallback(struct k_work *pxItem)
{
	ARG_UNUSED(pxItem);

	LOG_INF("OpenThread Connected");

	prvSubmitWork(&xStartCommissionerWork);
}

static void prvOtDisconnectCallback(struct k_work *pxItem)
{
	ARG_UNUSED(pxItem);

	LOG_INF("OpenThread Disconnected");
}

static void prvStartCommissioner(struct k_work *pxItem)
{
	otError xError;
	
	ARG_UNUSED(pxItem);

    // Start the Commissioner
    openthread_api_mutex_lock(openthread_get_default_context());
    xError = otCommissionerStart(openthread_get_default_instance(), prvOtCommissionerStateCallback, prvOtCommissionerJoinerCallback, openthread_get_default_instance());
    openthread_api_mutex_unlock(openthread_get_default_context());
    if (xError != OT_ERROR_NONE)
    {
        LOG_ERR("otCommissionerStart failed: %s", otThreadErrorToString(xError));
    }
	else
	{
		LOG_INF("Commissioner Started");
	}
}

static void prvAddJoiner(struct k_work *pxItem)
{
	otError xError;

	// Allow any joiner
	openthread_api_mutex_lock(openthread_get_default_context());
	xError = otCommissionerAddJoiner(openthread_get_default_instance(), NULL, THREAD_PASSPHRASE, THREAD_JOINER_TIMEOUT_MS);
	openthread_api_mutex_unlock(openthread_get_default_context());
	if (xError != OT_ERROR_NONE)
	{
		LOG_ERR("otCommissionerAddJoiner failed: %s", otThreadErrorToString(xError));
	}
	else
	{
		LOG_INF("Adding Joiner");
	}
}

void vThreadInit(void)
{
	static struct openthread_context *otContext;
	otError xError;

	otContext = openthread_get_default_context();

	if(otContext == NULL){
		LOG_ERR("openthread_get_default_context failed");
        return;
    }

	// otCoapSetDefaultHandler(srv_context.ot, coap_default_handler, NULL);
	// otCoapAddResource(srv_context.ot, &light_resource);
	// otCoapAddResource(srv_context.ot, &provisioning_resource);

	// xError = otCoapStart(srv_context.ot, COAP_PORT);
	// if (error != OT_ERROR_NONE) {
	// 	LOG_ERR("Failed to start OT CoAP. Error: %d", error);
	// 	goto end;
	// }


    openthread_set_state_changed_cb(prvOtStateChangeCallback);

	openthread_api_mutex_lock(otContext);

	// Enable IPv6 interface
	xError = otIp6SetEnabled(otContext->instance, true);
	if (xError != OT_ERROR_NONE)
	{
		LOG_ERR("otIp6SetEnabled failed: %s", otThreadErrorToString(xError));
	}

	// Channel
	xError = otLinkSetChannel(otContext->instance, THREAD_CHANNEL);
	if (xError != OT_ERROR_NONE)
	{
		LOG_ERR("otLinkSetChannel failed: %s", otThreadErrorToString(xError));
	}
	
	// PAN ID
	xError = otLinkSetPanId(otContext->instance, THREAD_PAN_ID);
	if (xError != OT_ERROR_NONE)
	{
		LOG_ERR("otLinkSetPanId failed: %s", otThreadErrorToString(xError));
	}

	// Network Name
	xError = otThreadSetNetworkName(otContext->instance, (char *)THREAD_NETWORK_NAME);
	if (xError != OT_ERROR_NONE)
	{
		LOG_ERR("otThreadSetNetworkName failed: %s", otThreadErrorToString(xError));
	}
	
	// Extended PAN ID
	static otExtendedPanId xPanId;
	if (net_bytes_from_str(xPanId.m8, OT_EXT_PAN_ID_SIZE, (char *)THREAD_EXT_PAN_ID) < 0)
	{
		LOG_ERR("net_bytes_from_str(xPanId) failed");
	}
	xError = otThreadSetExtendedPanId(otContext->instance, &xPanId);
	if (xError != OT_ERROR_NONE)
	{
		LOG_ERR("otThreadSetExtendedPanId failed: %s", otThreadErrorToString(xError));
	}

	// Network Key
	static otNetworkKey xNetworkKey;
	if (net_bytes_from_str(xNetworkKey.m8, OT_NETWORK_KEY_SIZE, (char *)THREAD_NETWORK_KEY) < 0)
	{
		LOG_ERR("net_bytes_from_str(xNetworkKey) failed");
	}
	xError = otThreadSetNetworkKey(otContext->instance, &xNetworkKey);
	if (xError != OT_ERROR_NONE)
	{
		LOG_ERR("otThreadSetNetworkKey failed: %s", otThreadErrorToString(xError));
	}

	// Start Thread
	xError = otThreadSetEnabled(otContext->instance, true);
	if (xError != OT_ERROR_NONE)
	{
		LOG_ERR("otThreadSetEnabled failed: %s", otThreadErrorToString(xError));
	}
}

